import argparse

from pathlib import Path
from datetime import datetime
from imlib.general.numerical import check_positive_float
from imlib.general.system import ensure_directory_exists
import imlib.IO.cells as cell_io

from imlib.general.exceptions import CommandLineInputError


def xml_scale(
    xml_file,
    x_scale=1,
    y_scale=1,
    z_scale=1,
    output_directory=None,
    integer=True,
):
    # TODO: add a csv option

    """
    To rescale the cell positions within an XML file. For compatibility with
    other software, or if  data has been scaled after cell detection.
    :param xml_file: Any cellfinder xml file
    :param x_scale: Rescaling factor in the first dimension
    :param y_scale: Rescaling factor in the second dimension
    :param z_scale: Rescaling factor in the third dimension
    :param output_directory: Directory to save the rescaled XML file.
    Defaults to the same directory as the input XML file
    :param integer: Force integer cell positions (default: True)
    :return:
    """
    if x_scale == y_scale == z_scale == 1:
        raise CommandLineInputError(
            "All rescaling factors are 1, " "please check the input."
        )
    else:
        input_file = Path(xml_file)
        start_time = datetime.now()
        cells = cell_io.get_cells(xml_file)

        for cell in cells:
            cell.transform(
                x_scale=x_scale,
                y_scale=y_scale,
                z_scale=z_scale,
                integer=integer,
            )

        if output_directory:
            output_directory = Path(output_directory)
        else:
            output_directory = input_file.parent

        ensure_directory_exists(output_directory)
        output_filename = output_directory / (input_file.stem + "_rescaled")
        output_filename = output_filename.with_suffix(input_file.suffix)

        cell_io.save_cells(cells, output_filename)

        print(
            "Finished. Total time taken: {}".format(
                datetime.now() - start_time
            )
        )


def get_parser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        dest="xml_file", type=str, help="XML file to be scaled"
    )

    parser.add_argument(
        "-x",
        "--x-scale",
        dest="x_scale",
        type=check_positive_float,
        default=1,
        help="Rescaling factor in the first dimension.",
    )
    parser.add_argument(
        "-y",
        "--y-scale",
        dest="y_scale",
        type=check_positive_float,
        default=1,
        help="Rescaling factor in the second dimension.",
    )
    parser.add_argument(
        "-z",
        "--z-scale",
        dest="z_scale",
        type=check_positive_float,
        default=1,
        help="Rescaling factor in the third dimension.",
    )
    parser.add_argument(
        "-o" "--output",
        dest="output_directory",
        type=str,
        default=None,
        help="Directory to save the rescaled XML file. Defaults to the same "
        "directory as the input XML file.",
    )
    return parser


def main():
    args = get_parser().parse_args()
    xml_scale(
        args.xml_file,
        x_scale=args.x_scale,
        y_scale=args.y_scale,
        z_scale=args.z_scale,
        output_directory=args.output_directory,
    )


if __name__ == "__main__":
    main()



import numpy as np
from tifffile import tifffile

from cellfinder.detect.filters.plane_filters.classical_filter import (
    enhance_peaks,
)
from cellfinder.detect.filters.plane_filters.tile_walker import TileWalker


class MpTileProcessor(object):
    def __init__(self, thread_q, ball_filter_q):
        self.thread_q = thread_q
        self.ball_filter_q = ball_filter_q

    def process(
        self,
        plane_id,
        path,
        previous_lock,
        self_lock,
        clipping_value,
        threshold_value,
        soma_diameter,
        log_sigma_size,
        n_sds_above_mean_thresh,
    ):
        laplace_gaussian_sigma = log_sigma_size * soma_diameter
        plane = tifffile.imread(path)
        plane = plane.T
        np.clip(plane, 0, clipping_value, out=plane)

        walker = TileWalker(plane, soma_diameter)

        walker.walk_out_of_brain_only()

        thresholded_img = enhance_peaks(
            walker.thresholded_img,
            clipping_value,
            gaussian_sigma=laplace_gaussian_sigma,
        )

        # threshold
        avg = thresholded_img.ravel().mean()
        sd = thresholded_img.ravel().std()

        plane[
            thresholded_img > avg + n_sds_above_mean_thresh * sd
        ] = threshold_value
        tile_mask = walker.good_tiles_mask.astype(np.uint8)

        with previous_lock:
            pass
        self.ball_filter_q.put((plane_id, plane, tile_mask))
        self.thread_q.put(plane_id)
        self_lock.release()



import os
import math
import logging

from tqdm import tqdm

from imlib.cells.cells import Cell
from imlib.IO.cells import save_cells
from tifffile import tifffile

from cellfinder.detect.filters.volume_filters.structure_detection import (
    get_structure_centre_wrapper,
)

from cellfinder.detect.filters.volume_filters.structure_splitting import (
    split_cells,
    StructureSplitException,
)
from cellfinder.detect.filters.setup_filters import setup


class Mp3DFilter(object):
    def __init__(
        self,
        data_queue,
        soma_diameter,
        output_folder,
        soma_size_spread_factor=1.4,
        setup_params=None,
        planes_paths_range=None,
        save_planes=False,
        plane_directory=None,
        start_plane=0,
        max_cluster_size=5000,
        outlier_keep=False,
        artifact_keep=True,
        save_csv=False,
    ):
        self.data_queue = data_queue
        self.soma_diameter = soma_diameter
        self.output_folder = output_folder
        self.soma_size_spread_factor = soma_size_spread_factor
        self.progress_bar = None
        self.planes_paths_range = planes_paths_range
        self.z = start_plane
        self.save_planes = save_planes
        self.plane_directory = plane_directory
        self.max_cluster_size = max_cluster_size
        self.outlier_keep = outlier_keep

        self.artifact_keep = artifact_keep

        self.output_file = "cells"
        self.save_csv = save_csv

        self.clipping_val = None
        self.threshold_value = None
        self.ball_filter = None
        self.cell_detector = None
        self.setup_params = setup_params

    def process(self):
        self.progress_bar = tqdm(
            total=len(self.planes_paths_range), desc="Processing planes"
        )

        self.ball_filter, self.cell_detector = setup(
            self.setup_params[0],
            self.setup_params[1],
            self.setup_params[2],
            self.setup_params[3],
            ball_overlap_fraction=self.setup_params[4],
            z_offset=self.setup_params[5],
        )

        while True:
            plane_id, plane, mask = self.data_queue.get()
            logging.debug(f"Plane {plane_id} received for 3D filtering")

            if plane_id is None:
                self.progress_bar.close()
                logging.debug("3D filter done")
                self.get_results()
                break

            logging.debug(f"Adding plane {plane_id} for 3D filtering")
            self.ball_filter.append(plane, mask)

            if self.ball_filter.ready:
                logging.debug(f"Ball filtering plane {plane_id}")
                self.ball_filter.walk()

                middle_plane = self.ball_filter.get_middle_plane()
                if self.save_planes:
                    self.save_plane(middle_plane)

                logging.debug(f"Detecting structures for plane {plane_id}")
                self.cell_detector.process(middle_plane)

                logging.debug(f"Structures done for plane {plane_id}")
                logging.debug(
                    f"Skipping plane {plane_id} for 3D filter"
                    " (out of bounds)"
                )

            self.z += 1
            if self.progress_bar is not None:
                self.progress_bar.update()

    def save_plane(self, plane):
        plane_name = f"plane_{str(self.z).zfill(4)}.tif"
        f_path = os.path.join(self.plane_directory, plane_name)
        tifffile.imsave(f_path, plane.T)

    def get_results(self):
        logging.info("Splitting cell clusters and writing results")

        max_cell_volume = sphere_volume(
            self.soma_size_spread_factor * self.soma_diameter / 2
        )

        cells = []
        for (
            cell_id,
            cell_points,
        ) in self.cell_detector.get_coords_list().items():
            cell_volume = len(cell_points)

            if cell_volume < max_cell_volume:
                cell_centre = get_structure_centre_wrapper(cell_points)
                cells.append(
                    Cell(
                        (cell_centre["x"], cell_centre["y"], cell_centre["z"]),
                        Cell.UNKNOWN,
                    )
                )
            else:
                if cell_volume < self.max_cluster_size:
                    try:
                        cell_centres = split_cells(
                            cell_points, outlier_keep=self.outlier_keep
                        )
                    except (ValueError, AssertionError) as err:
                        raise StructureSplitException(
                            f"Cell {cell_id}, error; {err}"
                        )
                    for cell_centre in cell_centres:
                        cells.append(
                            Cell(
                                (
                                    cell_centre["x"],
                                    cell_centre["y"],
                                    cell_centre["z"],
                                ),
                                Cell.UNKNOWN,
                            )
                        )
                else:
                    cell_centre = get_structure_centre_wrapper(cell_points)
                    cells.append(
                        Cell(
                            (
                                cell_centre["x"],
                                cell_centre["y"],
                                cell_centre["z"],
                            ),
                            Cell.ARTIFACT,
                        )
                    )

        xml_file_path = os.path.join(
            self.output_folder, self.output_file + ".xml"
        )
        save_cells(
            cells,
            xml_file_path,
            save_csv=self.save_csv,
            artifact_keep=self.artifact_keep,
        )


def sphere_volume(radius):
    return (4 / 3) * math.pi * radius ** 3



import math
import numpy as np


# WARNING: skewed if does not fall exactly (integer multiple)
def bin_array(arr, bin_width=15, bin_height=15):
    bin_width = int(bin_width)
    bin_height = int(bin_height)

    n_bins_width = int(math.ceil(arr.shape[0] / bin_width))
    width = n_bins_width * bin_width
    n_bins_height = int(math.ceil(arr.shape[1] / bin_height))
    height = n_bins_height * bin_height

    # zero padding array to integer multiples of bin_width/bin_height
    padded_array = np.zeros((width, height), dtype=arr.dtype)
    padded_array[0 : arr.shape[0], 0 : arr.shape[1]] = arr

    binned_array = padded_array.reshape(
        padded_array.shape[0] // bin_width,
        bin_width,
        padded_array.shape[1] // bin_height,
        bin_height,
    )
    return binned_array


def pad_and_bin_max(arr, bin_width=15, bin_height=15):
    binned_array = bin_array(arr, bin_width, bin_height)
    return binned_array.max(axis=(1, 3))


def pad_and_bin_mean(arr, bin_width=15, bin_height=15):
    binned_array = bin_array(arr, bin_width, bin_height)
    return binned_array.mean(axis=(1, 3))


def bin_mean_3d(arr, bin_width, bin_height, bin_depth):
    if (arr.shape[0] % bin_width) != 0:
        raise ValueError(
            "Bin width must divide arr exactly, got {} and {}".format(
                arr.shape[0], bin_width
            )
        )
    if (arr.shape[1] % bin_height) != 0:
        raise ValueError(
            "Bin height must divide arr exactly, got {} and {}".format(
                arr.shape[1], bin_height
            )
        )
    if (arr.shape[2] % bin_depth) != 0:
        raise ValueError(
            "Bin depth must divide arr exactly, got {} and {}".format(
                arr.shape[2], bin_depth
            )
        )

    binned_arr = []
    for i in range(0, arr.shape[2] - 1, bin_depth):
        sub_stack = [
            pad_and_bin_mean(arr[:, :, j], bin_width, bin_height)
            for j in range(i, i + bin_depth)
        ]
        binned_arr.append(np.dstack(sub_stack).mean(axis=2))
    binned_arr = np.dstack(binned_arr)
    return binned_arr



import numpy as np


def make_sphere(ball_shape, radius, position):
    """
    Assumes shape and position are both a 3-tuple of int or float
    the units are pixels / voxels (px for short)
    radius is a int or float in px

    :param tuple(int) ball_shape:
    :param float radius:
    :param tuple(int) position:
    :return:
    :rtype: np.ndarray
    """

    half_sizes = (radius,) * 3

    # generate the grid for the support points
    # centered at the position indicated by position
    grid = [slice(-x0, dim - x0) for x0, dim in zip(position, ball_shape)]
    position = np.ogrid[grid]
    # calculate the distance of all points from `position` center
    # scaled by the radius
    arr = np.zeros(ball_shape, dtype=float)
    for x_i, half_size in zip(position, half_sizes):
        arr += np.abs(x_i / half_size) ** 2
    # the inner part of the sphere will have distance below 1
    return arr <= 1.0


def four_connected_kernel():
    return np.array([[0, 1, 0], [1, 1, 1], [0, 1, 0]], dtype=np.bool)



import os
import sys
import pytest

from cellfinder.train.train_yml import main as train_run

data_dir = os.path.join(
    os.getcwd(), "tests", "data", "integration", "training"
)
cell_cubes = os.path.join(data_dir, "cells")
non_cell_cubes = os.path.join(data_dir, "non_cells")
training_yml_file = os.path.join(data_dir, "training.yml")


EPOCHS = "2"

# only checks that the model is trained, and then saved.
# doesn't check that it works etc


@pytest.mark.slow
def test_train(tmpdir):
    tmpdir = str(tmpdir)

    train_args = [
        "cellfinder_train",
        "-y",
        training_yml_file,
        "-o",
        tmpdir,
        "--epochs",
        EPOCHS,
    ]
    sys.argv = train_args
    train_run()

    model_file = os.path.join(tmpdir, "model.h5")
    assert os.path.exists(model_file)



import numpy as np

from scipy.ndimage import rotate, zoom

from cellfinder.tools.tools import (
    random_bool,
    random_probability,
    random_sign,
    all_elements_equal,
)

all_axes = np.array((0, 1, 2))


def augment(augmentation_parameters, image, scale_back=True):
    pixel_sizes = image.shape
    min_pixel_size = min(pixel_sizes)
    relative_pixel_sizes = []
    for pixel_size in pixel_sizes:
        relative_pixel_sizes.append(pixel_size / min_pixel_size)

    image, normalised_pixel_sizes = rescale_to_isotropic(
        image,
        relative_pixel_sizes,
        augmentation_parameters.interpolation_order,
    )
    # TODO: is this a sensible order?
    if augmentation_parameters.flip_axis is not None:
        image = flip_image(image, augmentation_parameters.axes_to_flip)

    if augmentation_parameters.translate is not None:
        image = translate_image(
            image,
            augmentation_parameters.translate_axes,
            augmentation_parameters.random_translate_multipliers,
        )

    # if augmentation_parameters.scale is not None:
    #     image = scale_image(image, augmentation_parameters.scale)

    if augmentation_parameters.rotate_max_axes is not None:
        image = rotate_image(image, augmentation_parameters.rotation_angles)

    if scale_back:
        image = rescale_to_original_size(
            image,
            relative_pixel_sizes,
            normalised_pixel_sizes,
            augmentation_parameters.interpolation_order,
        )
    return image


def rescale_to_isotropic(image, relative_pixel_sizes, interpolation_order):

    if not all_elements_equal(relative_pixel_sizes):
        min_pixel_size = min(relative_pixel_sizes)
        normalised_pixel_sizes = []
        for pixel_size in relative_pixel_sizes:
            normalised_pixel_sizes.append(
                round(pixel_size / min_pixel_size, 2)
            )

        image = zoom(image, normalised_pixel_sizes, order=interpolation_order)
    else:
        normalised_pixel_sizes = relative_pixel_sizes
    return image, normalised_pixel_sizes


def rescale_to_original_size(
    image, relative_pixel_sizes, normalised_pixel_sizes, interpolation_order
):
    if not all_elements_equal(relative_pixel_sizes):
        inverse_pixel_sizes = []
        for pixel_size in normalised_pixel_sizes:
            inverse_pixel_sizes.append(round(1 / pixel_size, 2))

        image = zoom(image, inverse_pixel_sizes, order=interpolation_order)
    return image


def flip_image(image, axes_to_flip):
    for axis in axes_to_flip:
        image = np.flip(image, axis)
    return image


def translate_image(image, translate_axes, random_translate_multipliers):
    pixel_shifts = []
    for idx, axis in enumerate(translate_axes):
        pixel_shifts.append(
            int(round(random_translate_multipliers[idx] * image.shape[axis]))
        )

    image = np.roll(image, pixel_shifts, axis=translate_axes)
    return image


def rotate_image(image, rotation_angles):
    for axis, angle in enumerate(rotation_angles):
        if angle != 0:
            rotate_axes = all_axes[all_axes != axis]
            image = rotate(
                image, angle, axes=rotate_axes, reshape=False, mode="constant"
            )
    return image


# def scale_image(image, scale, ndigits=2):
#     scale_factor = round(
#         uniform(scale[0], scale[1]), ndigits=ndigits
#     )
#     return image

# def shear_image(image):
#     return image


class AugmentationParameters:
    # precomputed, so both channels are treated identically
    def __init__(
        self,
        flip_axis,
        translate,
        rotate_max_axes,
        interpolation_order,
        augment_likelihood,
    ):
        # this is a clumsy way of passing parameters to the augment function
        self.flip_axis = flip_axis
        self.translate = translate
        self.rotate_max_axes = rotate_max_axes
        self.interpolation_order = interpolation_order

        self.augment_likelihood = augment_likelihood

        self.axes_to_flip = None
        self.translate_axes = None
        self.random_translate_multipliers = None
        self.rotation_angles = None

        if flip_axis:
            self.get_flip_parameters(flip_axis)
        if translate:
            self.get_translation_parameters(translate)
        if rotate_max_axes:
            self.get_rotation_parameters(rotate_max_axes)

    def get_flip_parameters(self, flip_axis):
        self.axes_to_flip = []
        for axis in all_axes:
            if axis in flip_axis:
                if random_bool(likelihood=self.augment_likelihood):
                    self.axes_to_flip.append(axis)

    def get_translation_parameters(self, translate):
        self.translate_axes = []
        self.random_translate_multipliers = []
        for axis, translate_mag in enumerate(translate):
            if translate_mag > 0:
                if random_bool(likelihood=self.augment_likelihood):
                    self.translate_axes.append(axis)
                    self.random_translate_multipliers.append(
                        random_sign() * random_probability() * translate_mag
                    )

    def get_rotation_parameters(self, rotate_max_axes):
        self.rotation_angles = []
        for max_rotation in rotate_max_axes:
            if random_bool(likelihood=self.augment_likelihood):
                angle = int(
                    round(
                        -max_rotation + 2 * random_probability() * max_rotation
                    )
                )
            else:
                angle = 0
            self.rotation_angles.append(angle)




# deprecated cimport for backwards compatibility:
from libc.string cimport const_char


cdef extern from "<string>" namespace "std" nogil:

    size_t npos = -1

    cdef cppclass string:
        cppclass iterator:
            iterator()
            char& operator*()
            iterator(iterator &)
            iterator operator++()
            iterator operator--()
            bint operator==(iterator)
            bint operator!=(iterator)
        cppclass reverse_iterator:
            char& operator*()
            iterator operator++()
            iterator operator--()
            iterator operator+(size_t)
            iterator operator-(size_t)
            bint operator==(reverse_iterator)
            bint operator!=(reverse_iterator)
            bint operator<(reverse_iterator)
            bint operator>(reverse_iterator)
            bint operator<=(reverse_iterator)
            bint operator>=(reverse_iterator)
        cppclass const_iterator(iterator):
            pass
        cppclass const_reverse_iterator(reverse_iterator):
            pass

        string() except +
        string(const char *) except +
        string(const char *, size_t) except +
        string(const string&) except +
        # as a string formed by a repetition of character c, n times.
        string(size_t, char) except +
        # from a pair of iterators
        string(iterator first, iterator last) except +

        iterator begin()
        const_iterator const_begin "begin"()
        iterator end()
        const_iterator const_end "end"()
        reverse_iterator rbegin()
        const_reverse_iterator const_rbegin "rbegin"()
        reverse_iterator rend()
        const_reverse_iterator const_rend "rend"()

        const char* c_str()
        const char* data()
        size_t size()
        size_t max_size()
        size_t length()
        void resize(size_t)
        void resize(size_t, char c)
        size_t capacity()
        void reserve(size_t)
        void clear()
        bint empty()
        iterator erase(iterator position)
        iterator erase(const_iterator position)
        iterator erase(iterator first, iterator last)
        iterator erase(const_iterator first, const_iterator last)

        char& at(size_t)
        char& operator[](size_t)
        char& front()  # C++11
        char& back()   # C++11
        int compare(const string&)

        string& append(const string&)
        string& append(const string&, size_t, size_t)
        string& append(const char *)
        string& append(const char *, size_t)
        string& append(size_t, char)

        void push_back(char c)

        string& assign (const string&)
        string& assign (const string&, size_t, size_t)
        string& assign (const char *, size_t)
        string& assign (const char *)
        string& assign (size_t n, char c)

        string& insert(size_t, const string&)
        string& insert(size_t, const string&, size_t, size_t)
        string& insert(size_t, const char* s, size_t)


        string& insert(size_t, const char* s)
        string& insert(size_t, size_t, char c)

        size_t copy(char *, size_t, size_t)

        size_t find(const string&)
        size_t find(const string&, size_t)
        size_t find(const char*, size_t pos, size_t)
        size_t find(const char*, size_t pos)
        size_t find(char, size_t pos)

        size_t rfind(const string&, size_t)
        size_t rfind(const char* s, size_t, size_t)
        size_t rfind(const char*, size_t pos)
        size_t rfind(char c, size_t)
        size_t rfind(char c)

        size_t find_first_of(const string&, size_t)
        size_t find_first_of(const char* s, size_t, size_t)
        size_t find_first_of(const char*, size_t pos)
        size_t find_first_of(char c, size_t)
        size_t find_first_of(char c)

        size_t find_first_not_of(const string&, size_t)
        size_t find_first_not_of(const char* s, size_t, size_t)
        size_t find_first_not_of(const char*, size_t pos)
        size_t find_first_not_of(char c, size_t)
        size_t find_first_not_of(char c)

        size_t find_last_of(const string&, size_t)
        size_t find_last_of(const char* s, size_t, size_t)
        size_t find_last_of(const char*, size_t pos)
        size_t find_last_of(char c, size_t)
        size_t find_last_of(char c)

        size_t find_last_not_of(const string&, size_t)
        size_t find_last_not_of(const char* s, size_t, size_t)
        size_t find_last_not_of(const char*, size_t pos)

        string substr(size_t, size_t)
        string substr()
        string substr(size_t)

        size_t find_last_not_of(char c, size_t)
        size_t find_last_not_of(char c)

        #string& operator= (const string&)
        #string& operator= (const char*)
        #string& operator= (char)

        string operator+ (const string& rhs)
        string operator+ (const char* rhs)

        bint operator==(const string&)
        bint operator==(const char*)

        bint operator!= (const string& rhs )
        bint operator!= (const char* )

        bint operator< (const string&)
        bint operator< (const char*)

        bint operator> (const string&)
        bint operator> (const char*)

        bint operator<= (const string&)
        bint operator<= (const char*)

        bint operator>= (const string&)
        bint operator>= (const char*)



from pathlib import Path

from imlib.general.exceptions import CommandLineInputError


def get_subdirectories(directory, names_only=False):
    """
    Return all subdirectories in a given directory
    :param directory:
    :param names_only: If true, dont return paths, but the names
    :return: Subdirectories
    """

    p = Path(directory)
    if names_only:
        return [x.name for x in p.iterdir() if x.is_dir()]
    else:
        return [x for x in p.iterdir() if x.is_dir()]


def get_number_of_files_in_dir(directory):
    """
    Sums the number of files in a directory
    :param directory: Any directory with files
    :return: Number of files in directory
    """
    directory = Path(directory)
    files = directory.iterdir()
    total_files = sum(1 for x in files)
    return total_files


def check_path_exists(file):
    """
    Returns True is a file exists, otherwise throws a FileNotFoundError
    :param file: Input file
    :return: True, if the file exists
    """
    file = Path(file)
    if file.exists():
        return True
    else:
        raise FileNotFoundError


def catch_input_file_error(path):
    """
    Catches if an input path doesn't exist, and returns an informative error
    :param path: Input file path
    default)
    """
    try:
        check_path_exists(path)
    except FileNotFoundError:
        message = (
            "File path: '{}' cannot be found. Please check your input "
            "arguments.".format(path)
        )
        raise CommandLineInputError(message)


def memory_in_bytes(memory_amount, unit):
    """
    Converts an amount of memory (in given units) to bytes

    :param float memory_amount: An amount of memory
    :param str unit: The units ('KB', 'MB', 'GB', 'TB', 'PB')
    :return: Amount of memory in bytes
    """
    if memory_amount is None:
        return memory_amount

    supported_units = {"KB": 3, "MB": 6, "GB": 9, "TB": 12, "PB": 15}

    unit = unit.upper()
    if unit not in supported_units:
        raise NotImplementedError(
            f"Unit: {unit} is not supported. Please "
            f"use one of {list(supported_units.keys())}"
        )
    else:
        return memory_amount * 10 ** supported_units[unit]



import logging

from skimage.io import imread
import pandas as pd
import numpy as np
from tqdm import tqdm
import tensorflow as tf

from cellfinder.classify.resnet import build_model


def get_model(
    existing_model=None,
    model_weights=None,
    network_depth=None,
    learning_rate=0.0001,
    inference=False,
    continue_training=False,
):
    """
    Returns the correct model based on the arguments passed
    :param existing_model: An existing, trained model. This is returned if it
    exists
    :param model_weights: This file is used to set the model weights if it
    exists
    :param network_depth: This defines the type of model to be created if
    necessary
    :param learning_rate: For creating a new model
    :param inference: If True, will ensure that a trained model exists. E.g.
    by using the default one
    :param continue_training: If True, will ensure that a trained model
    exists. E.g. by using the default one
    :return: A tf.keras model

    """
    if existing_model is not None:
        logging.debug(f"Loading model: {existing_model}")
        return tf.keras.models.load_model(existing_model)
    else:
        logging.debug(f"Creating a new instance of model: {network_depth}")
        model = build_model(
            network_depth=network_depth, learning_rate=learning_rate
        )
        if inference or continue_training:
            logging.debug(
                f"Setting model weights according to: {model_weights}"
            )
            model.load_weights(model_weights)
        return model


def populate_array_with_cubes(images, idx, signal_im, background_im):
    # if paths are pathlib objs, skimage only reads one plane
    images[idx, :, :, :, 0] = np.moveaxis(imread(signal_im), 0, 2)
    images[idx, :, :, :, 1] = np.moveaxis(imread(background_im), 0, 2)
    return images


def load_cubes_from_dir(
    directory, shape, signal_extension, background_extension
):

    cubes_list = list(directory.glob("*" + signal_extension))
    number_images = len(cubes_list)
    images = np.empty(((number_images,) + shape))

    for idx, signal_im in enumerate(cubes_list):
        signal_im = str(signal_im)
        background_im = signal_im.replace(
            signal_extension, background_extension
        )
        images = populate_array_with_cubes(
            images, idx, signal_im, background_im
        )

    return images.astype(np.float16)


def load_cubes_from_csv(
    signal_csv, background_csv, labels_csv, shape, n_test_samples, skip_amt
):
    signal = pd.read_csv(signal_csv, header=None)
    signal_list = list(signal[1])
    if n_test_samples is not None:
        signal_list = signal_list[0:n_test_samples:skip_amt]

    background = pd.read_csv(background_csv, header=None)
    background_list = list(background[1])
    if n_test_samples is not None:
        background_list = background_list[0:n_test_samples:skip_amt]

    number_images = len(signal_list)
    images = np.empty(((number_images,) + shape))

    print(f"Loading {number_images} images")
    for idx, signal_im in enumerate(tqdm(signal_list)):
        background_im = background_list[idx]
        images = populate_array_with_cubes(
            images, idx, signal_im, background_im
        )

    # required editing yaml handling
    labels = pd.read_csv(labels_csv, header=None)[1]
    labels = np.array(labels).astype(np.int8)
    if n_test_samples is not None:
        labels = labels[0:n_test_samples:skip_amt]
    labels = tf.keras.utils.to_categorical(labels).astype(np.uint8)
    return images, labels


def load_data(
    shape, cells_dir, no_cells_dir, signal_extension, background_extension
):
    cell_images = load_cubes_from_dir(
        cells_dir, shape, signal_extension, background_extension
    )
    non_cell_images = load_cubes_from_dir(
        no_cells_dir, shape, signal_extension, background_extension
    )
    images = np.append(cell_images, non_cell_images, axis=0)

    cell_labels = np.ones((len(cell_images), 1))
    non_cell_labels = np.zeros((len(non_cell_images), 1))
    labels = np.append(cell_labels, non_cell_labels)
    labels = tf.keras.utils.to_categorical(labels).astype(np.uint8)

    return images, labels


def make_lists(tiff_files, train=True):

    signal_list = []
    background_list = []
    if train:
        labels = []

    for group in tiff_files:
        for image in group:
            signal_list.append(image.img_files[0])
            background_list.append(image.img_files[1])
            if train:
                if image.label == "no_cell":
                    labels.append(0)
                elif image.label == "cell":
                    labels.append(1)

    if train:
        labels = np.array(labels)
        return signal_list, background_list, labels
    else:
        return signal_list, background_list


def batch_viewer(
    image_batch,
    opacity=0.5,
    signal_name="Signal channnel",
    background_name="Background channel",
    signal_colormap="magenta",
    background_colormap="cyan",
):
    """
    Tool to visualise cube generator batches
    :param image_batch:
    :param opacity:
    :param signal_name:
    :param background_name:
    :param signal_colormap:
    :param background_colormap:
    :return:
    """
    import napari

    image = np.moveaxis(image_batch, 3, 1)
    scale = calculate_scale(image[0, :, :, :, 0])

    with napari.gui_qt():
        v = napari.Viewer()
        v.add_image(
            image[..., 0],
            name=signal_name,
            opacity=opacity,
            colormap=signal_colormap,
            scale=scale,
        )
        v.add_image(
            image[..., 1],
            name=background_name,
            opacity=opacity,
            colormap=background_colormap,
            scale=scale,
        )


def calculate_scale(image):
    scale = []
    for dim in image.shape:
        scale.append(100 / dim)
    return tuple(scale)



import logging
import numpy as np
from imlib.general.system import get_sorted_file_paths, get_num_processes


from imlib.IO.cells import save_cells
from cellfinder.classify.tools import get_model
from cellfinder.classify.cube_generator import CubeGeneratorFromFile
from cellfinder.train.train_yml import models


def main(args, max_workers=3):
    signal_paths = args.signal_planes_paths[args.signal_channel]
    background_paths = args.background_planes_path[0]
    signal_images = get_sorted_file_paths(signal_paths, file_extension="tif")
    background_images = get_sorted_file_paths(
        background_paths, file_extension="tif"
    )

    # Too many workers doesn't increase speed, and uses huge amounts of RAM
    workers = get_num_processes(
        min_free_cpu_cores=args.n_free_cpus, n_max_processes=max_workers
    )

    logging.debug("Initialising cube generator")
    inference_generator = CubeGeneratorFromFile(
        args.paths.cells_file_path,
        signal_images,
        background_images,
        batch_size=args.batch_size,
        x_pixel_um=args.x_pixel_um,
        y_pixel_um=args.y_pixel_um,
        z_pixel_um=args.z_pixel_um,
        x_pixel_um_network=args.x_pixel_um_network,
        y_pixel_um_network=args.y_pixel_um_network,
        z_pixel_um_network=args.z_pixel_um_network,
        cube_width=args.cube_width,
        cube_height=args.cube_height,
        cube_depth=args.cube_depth,
    )

    model = get_model(
        existing_model=args.trained_model,
        model_weights=args.model_weights,
        network_depth=models[args.network_depth],
        inference=True,
    )

    logging.info("Running inference")
    predictions = model.predict(
        inference_generator,
        use_multiprocessing=True,
        workers=workers,
        verbose=True,
    )
    predictions = predictions.round()
    predictions = predictions.astype("uint16")

    predictions = np.argmax(predictions, axis=1)
    cells_list = []

    # only go through the "extractable" cells
    for idx, cell in enumerate(inference_generator.ordered_cells):
        cell.type = predictions[idx] + 1
        cells_list.append(cell)

    logging.info("Saving classified cells")
    save_cells(
        cells_list, args.paths.classification_out_file, save_csv=args.save_csv
    )



import argparse
import os
import logging

import pandas as pd
from tqdm import tqdm


from brainio import brainio

from imlib.IO.cells import get_cells
from imlib.pandas.misc import sanitise_df
from imlib.image.metadata import define_pixel_sizes
from imlib.general.config import get_config_obj
from imlib.source.source_files import get_structures_path

from neuro.structures.structures_tree import (
    atlas_value_to_structure_id,
    CellCountMissingCellsException,
    UnknownAtlasValue,
)
from neuro.atlas_tools.misc import get_atlas_pixel_sizes
from neuro.structures.IO import load_structures_as_df

import cellfinder.tools.parser as cellfinder_parse
from cellfinder.tools.prep import prep_atlas_conf, Paths

LEFT_HEMISPHERE = 2
RIGHT_HEMISPHERE = 1


def region_summary_cli_parser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser = cli_parse(parser)
    parser = cellfinder_parse.count_summary_parse(parser)
    parser = cellfinder_parse.pixel_parser(parser)
    return parser


def cli_parse(parser):
    cli_parser = parser.add_argument_group("Input data options")
    cli_parser.add_argument(
        "--registered-atlas",
        dest="registered_atlas_path",
        type=str,
        help="The path to the atlas registered to the sample brain",
    )
    cli_parser.add_argument(
        "--hemispheres",
        dest="hemispheres_atlas_path",
        type=str,
        help="The atlas with just the hemispheres encoded.",
    )
    cli_parser.add_argument(
        "--xml",
        dest="xml_file_path",
        type=str,
        help="The xml file containing the cell locations",
    )

    cli_parser.add_argument(
        "-o",
        "--output-dir",
        dest="output_dir",
        type=str,
        required=True,
        help="Output directory for all intermediate and final results.",
    )

    return parser


def get_cells_data(xml_file_path, cells_only=True):
    cells = get_cells(xml_file_path, cells_only=cells_only)
    if not cells:
        raise CellCountMissingCellsException(
            "No cells found in file: {}".format(xml_file_path)
        )
    return cells


def get_scales(sample_pixel_sizes, atlas_pixel_sizes, scale=True):
    if scale:
        sample_x, sample_y, sample_z = sample_pixel_sizes
        atlas_x_pix_size = atlas_pixel_sizes["x"]
        atlas_y_pix_size = atlas_pixel_sizes["y"]
        atlas_z_pix_size = atlas_pixel_sizes["z"]

        x_scale = float(sample_x) / float(atlas_x_pix_size)
        y_scale = float(sample_y) / float(atlas_y_pix_size)
        z_scale = float(sample_z) / float(atlas_z_pix_size)
        return x_scale, y_scale, z_scale
    else:
        return 1, 1, 1


def get_atlas_pixel_sizes(atlas_config_path):
    config_obj = get_config_obj(atlas_config_path)
    atlas_conf = config_obj["atlas"]
    atlas_pixel_sizes = atlas_conf["pixel_size"]
    return atlas_pixel_sizes


def get_max_coords(cells):
    max_x, max_y, max_z = (0, 0, 0)
    for cell in cells:
        if cell.x > max_x:
            max_x = cell.x
        if cell.y > max_y:
            max_y = cell.y
        if cell.z > max_z:
            max_z = cell.z
    return max_x, max_y, max_z


def transform_cell_coords(atlas, cell, scales):
    x_scale, y_scale, z_scale = scales
    # convert to atlas coordinates
    cell.soft_transform(x_scale, y_scale, z_scale, integer=True)
    # In case we reorientated the sample, not the atlas.
    flip_dims = False  # FIXME: put CLI option
    if flip_dims:
        cell.transformed_y, cell.transformed_z = (
            cell.transformed_z,
            cell.transformed_y,
        )  # TODO: do in cell
        # TEST: check that correct dim to flip
        cell.transformed_z = atlas.shape[2] - cell.transformed_z


def get_cells_nbs_df(cells, structures_reference_df, structures_with_cells):
    structures_with_cells = list(structures_with_cells)
    cell_numbers = pd.DataFrame(
        columns=("structure_name", "hemisphere", "cell_count")
    )
    for structure in structures_with_cells:
        for hemisphere in (1, 2):
            n_cells = len(
                [
                    c
                    for c in cells
                    if c.structure_id == structure
                    and c.hemisphere == hemisphere
                ]
            )
            if n_cells:
                struct_name = structures_reference_df[
                    structures_reference_df["structure_id_path"] == structure
                ]["name"].values[0]
                cell_numbers = cell_numbers.append(
                    {
                        "structure_name": struct_name,
                        "hemisphere": hemisphere,
                        "cell_count": n_cells,
                    },
                    ignore_index=True,
                )
    sorted_cell_numbers = cell_numbers.sort_values(
        by=["cell_count"], ascending=False
    )
    return sorted_cell_numbers


def get_structure_from_coordinates(
    atlas, cell, max_coords, order=(0, 1, 2), structures_reference_df=None
):
    transformed_coords = (
        cell.transformed_x,
        cell.transformed_y,
        cell.transformed_z,
    )
    try:
        atlas_value = atlas[
            transformed_coords[order[0]],
            transformed_coords[order[1]],
            transformed_coords[order[2]],
        ]
    except IndexError as err:
        logging.warning(
            "The cell {}, scaled to {} "
            "falls outside of the atlas with "
            "dimensions {}. Treating as outside the brain".format(
                cell, transformed_coords, atlas.shape, max_coords, err
            )
        )
        return 0

    if structures_reference_df is None:
        return atlas_value
    else:
        try:
            structure_id = atlas_value_to_structure_id(
                atlas_value, structures_reference_df
            )
        except UnknownAtlasValue as err:
            print(
                "Skipping cell {} (scaled: {}), missing value {}".format(
                    cell, transformed_coords, err
                )
            )
            return
        else:
            return structure_id


def analysis_run(args, file_name="summary_cell_counts.csv"):
    args = prep_atlas_conf(args)

    atlas = brainio.load_any(args.paths.registered_atlas_path)
    hemisphere = brainio.load_any(args.paths.hemispheres_atlas_path)

    cells = get_cells_data(
        args.paths.classification_out_file, cells_only=args.cells_only,
    )
    max_coords = get_max_coords(cells)  # Useful for debugging dimensions
    structures_reference_df = load_structures_as_df(get_structures_path())

    atlas_pixel_sizes = get_atlas_pixel_sizes(args.atlas_config)
    sample_pixel_sizes = args.x_pixel_um, args.y_pixel_um, args.z_pixel_um

    scales = get_scales(
        sample_pixel_sizes, atlas_pixel_sizes, args.scale_cell_coordinates
    )

    structures_with_cells = set()
    for i, cell in enumerate(tqdm(cells)):
        transform_cell_coords(atlas, cell, scales)

        structure_id = get_structure_from_coordinates(
            atlas,
            cell,
            max_coords,
            order=args.coordinates_order,
            structures_reference_df=structures_reference_df,
        )
        if structure_id is not None:
            cell.structure_id = structure_id

            structures_with_cells.add(structure_id)
        else:
            continue

        cell.hemisphere = get_structure_from_coordinates(
            hemisphere, cell, max_coords, order=args.coordinates_order
        )

    sorted_cell_numbers = get_cells_nbs_df(
        cells, structures_reference_df, structures_with_cells
    )

    combined_hemispheres = combine_df_hemispheres(sorted_cell_numbers)
    df = calculate_densities(combined_hemispheres, args.paths.volume_csv_path)
    df = sanitise_df(df)
    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir)
    output_file = os.path.join(args.output_dir, file_name)
    df.to_csv(output_file, index=False)


def calculate_densities(counts, volume_csv_path):
    """
    Use the region volume information from registration to calculate cell
    densities. Based on the atlas names, which must be exactly equal.
    :param counts: dataframe with cell counts
    :param volume_csv_path: path of the volumes of each brain region
    :return:
    """
    volumes = pd.read_csv(volume_csv_path, sep=",", header=0, quotechar='"')
    df = pd.merge(counts, volumes, on="structure_name", how="outer")
    df = df.fillna(0)
    df["left_cells_per_mm3"] = df.left_cell_count / df.left_volume_mm3
    df["right_cells_per_mm3"] = df.right_cell_count / df.right_volume_mm3
    return df


def combine_df_hemispheres(df):
    """
    Combine left and right hemisphere data onto a single row
    :param df:
    :return:
    """
    left = df[df["hemisphere"] == LEFT_HEMISPHERE]
    right = df[df["hemisphere"] == RIGHT_HEMISPHERE]
    left = left.drop(["hemisphere"], axis=1)
    right = right.drop(["hemisphere"], axis=1)
    left.rename(columns={"cell_count": "left_cell_count"}, inplace=True)
    right.rename(columns={"cell_count": "right_cell_count"}, inplace=True)
    both = pd.merge(left, right, on="structure_name", how="outer")
    both = both.fillna(0)
    both["total_cells"] = both.left_cell_count + both.right_cell_count
    both = both.sort_values("total_cells", ascending=False)
    return both


def main():
    args = region_summary_cli_parser().parse_args()
    args = define_pixel_sizes(args)
    args.paths = Paths(args, args.output_dir)
    args.paths.registered_atlas_path = args.registered_atlas_path
    args.paths.hemispheres_atlas_path = args.hemispheres_atlas_path
    args.paths.classification_out_file = args.xml_file_path
    analysis_run(args)


if __name__ == "__main__":
    main()



import pytest

import numpy as np

from cellfinder.detect.filters.volume_filters import structure_detection


def test_get_non_zero_ull_min():
    assert (
        structure_detection.get_non_zero_ull_min_wrapper(list(range(10))) == 1
    )
    assert (
        structure_detection.get_non_zero_ull_min_wrapper([0] * 10)
        == (2 ** 64) - 1
    )


class Point:
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z

    def __str__(self):
        return "x: {}, y: {}, z: {}".format(self.x, self.y, self.z)


@pytest.fixture()
def three_d_cross():
    return np.array(
        [
            [[0, 0, 0], [0, 1, 0], [0, 0, 0]],
            [[0, 1, 0], [1, 1, 1], [0, 1, 0]],
            [[0, 0, 0], [0, 1, 0], [0, 0, 0]],
        ]
    )


@pytest.fixture()
def structure(three_d_cross):
    coords = np.where(three_d_cross == 1)[::-1]
    s = [Point(*c) for c in zip(coords[0], coords[1], coords[2])]
    return s


def test_get_structure_centre(structure):
    result_point = structure_detection.get_structure_centre_wrapper(structure)
    assert (result_point["x"], result_point["y"], result_point["z"]) == (
        1,
        1,
        1,
    )



cdef extern from "<numeric>" namespace "std" nogil:
    T inner_product[InputIt1, InputIt2, T](InputIt1 first1, InputIt1 last1, InputIt2 first2, T init)

    T inner_product[InputIt1, InputIt2, T, BinaryOperation1, BinaryOperation2](InputIt1 first1, InputIt1 last1,
                                                                               InputIt2 first2, T init,
                                                                               BinaryOperation1 op1,
                                                                               BinaryOperation2 op2)

    void iota[ForwardIt, T](ForwardIt first, ForwardIt last, T value)

    T accumulate[InputIt, T](InputIt first, InputIt last, T init)

    T accumulate[InputIt, T, BinaryOperation](InputIt first, InputIt last, T init, BinaryOperation op)

    void adjacent_difference[InputIt, OutputIt](InputIt in_first, InputIt in_last, OutputIt out_first)

    void adjacent_difference[InputIt, OutputIt, BinaryOperation](InputIt in_first, InputIt in_last, OutputIt out_first,
                                                                 BinaryOperation op)

    void partial_sum[InputIt, OutputIt](InputIt in_first, OutputIt in_last, OutputIt out_first)

    void partial_sum[InputIt, OutputIt, BinaryOperation](InputIt in_first, InputIt in_last, OutputIt out_first,
                                                         BinaryOperation op)



"""
main
===============

Trains a network based on a yaml file specifying cubes of cells/non cells.

N.B imports are within functions to prevent tensorflow being imported before
it's warnings are silenced
"""


import os

from datetime import datetime
from pathlib import Path
from argparse import (
    ArgumentParser,
    ArgumentDefaultsHelpFormatter,
    ArgumentTypeError,
)
from sklearn.model_selection import train_test_split
from imlib.general.numerical import check_positive_float, check_positive_int
from imlib.general.system import ensure_directory_exists
from imlib.IO.cells import find_relevant_tiffs
from imlib.IO.yaml import read_yaml_section

tf_suppress_log_messages = [
    "sample_weight modes were coerced from",
    "multiprocessing can interact badly with TensorFlow",
]

models = {
    "18": "18-layer",
    "34": "34-layer",
    "50": "50-layer",
    "101": "101-layer",
    "152": "152-layer",
}


def valid_model_depth(depth):
    """
    Ensures a correct existing_model is chosen
    :param value: Input value
    :param models: Dict of allowed models
    :return: Input value, if it corresponds to a valid existing_model
    """

    if depth in models.keys():
        return depth
    else:
        raise ArgumentTypeError(
            f"Model depth: {depth} is not valid. Please "
            f"choose one of: {list(models.keys())}"
        )


def training_parse():
    from cellfinder.tools.parser import misc_parse
    from cellfinder.download.cli import model_parser, download_directory_parser

    training_parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter
    )
    training_parser.add_argument(
        "-y",
        "--yaml",
        dest="yaml_file",
        nargs="+",
        required=True,
        type=str,
        help="The path to the yaml run file.",
    )
    training_parser.add_argument(
        "-o",
        "--output-dir",
        dest="output_dir",
        required=True,
        type=str,
        help="Output directory for the final model.",
    )
    training_parser.add_argument(
        "--continue-training",
        dest="continue_training",
        action="store_true",
        help="Continue training from an existing trained model. If no model "
        "or model weights are specified, this will continue from the "
        "included model.",
    )
    training_parser.add_argument(
        "--trained-model",
        dest="trained_model",
        type=str,
        help="Path to the trained model",
    )
    training_parser.add_argument(
        "--model-weights",
        dest="model_weights",
        type=str,
        help="Path to existing model weights",
    )
    training_parser.add_argument(
        "--network-depth",
        dest="network_depth",
        type=valid_model_depth,
        default="50",
        help="Resnet depth (based on He et al. (2015)",
    )
    training_parser.add_argument(
        "--batch-size",
        dest="batch_size",
        type=check_positive_int,
        default=16,
        help="Training batch size",
    )
    training_parser.add_argument(
        "--epochs",
        dest="epochs",
        type=check_positive_int,
        default=100,
        help="Number of training epochs",
    )
    training_parser.add_argument(
        "--test-fraction",
        dest="test_fraction",
        type=float,
        default=0.1,
        help="Fraction of training data to use for validation",
    )
    training_parser.add_argument(
        "--learning-rate",
        dest="learning_rate",
        type=check_positive_float,
        default=0.0001,
        help="Learning rate for training the model",
    )
    training_parser.add_argument(
        "--no-augment",
        dest="no_augment",
        action="store_true",
        help="Don't apply data augmentation",
    )
    training_parser.add_argument(
        "--save-weights",
        dest="save_weights",
        action="store_true",
        help="Only store the model weights, and not the full model. Useful to "
        "save storage space.",
    )
    training_parser.add_argument(
        "--save-checkpoints",
        dest="save_checkpoints",
        action="store_true",
        help="Store the model at intermediate points during training",
    )
    training_parser.add_argument(
        "--checkpoint-interval",
        dest="checkpoint_interval",
        type=check_positive_int,
        default=1,
        help="Number of epochs between checkpoints",
    )
    training_parser.add_argument(
        "--tensorboard",
        action="store_true",
        help="Log to output_directory/tensorboard",
    )

    training_parser = misc_parse(training_parser)
    training_parser = model_parser(training_parser)
    training_parser = download_directory_parser(training_parser)
    args = training_parser.parse_args()

    return args


def parse_yaml(yaml_files, section="data"):
    data = []
    for yaml_file in yaml_files:
        data.extend(read_yaml_section(yaml_file, section))
    return data


def get_tiff_files(yaml_contents):
    from cellfinder.tools.tiff import TiffList, TiffDir

    tiff_lists = []
    for d in yaml_contents:
        if d["bg_channel"] < 0:
            channels = [d["signal_channel"]]
        else:
            channels = [d["signal_channel"], d["bg_channel"]]
        if "cell_def" in d and d["cell_def"]:
            ch1_tiffs = [
                os.path.join(d["cube_dir"], f)
                for f in os.listdir(d["cube_dir"])
                if f.endswith("Ch" + str(channels[0]) + ".tif")
            ]
            tiff_lists.append(
                TiffList(
                    find_relevant_tiffs(ch1_tiffs, d["cell_def"]),
                    channels,
                    d["type"],
                )
            )
        else:
            tiff_lists.append(TiffDir(d["cube_dir"], channels, d["type"]))

    tiff_files = [tiff_dir.make_tifffile_list() for tiff_dir in tiff_lists]
    return tiff_files


def main():
    from cellfinder.main import suppress_tf_logging

    suppress_tf_logging(tf_suppress_log_messages)

    from tensorflow.keras.callbacks import TensorBoard, ModelCheckpoint

    from cellfinder.tools.prep import prep_training
    from cellfinder.classify.tools import make_lists, get_model
    from cellfinder.classify.cube_generator import CubeGeneratorFromDisk

    start_time = datetime.now()
    args = training_parse()
    output_dir = Path(args.output_dir)
    ensure_directory_exists(output_dir)
    args = prep_training(args)
    yaml_contents = parse_yaml(args.yaml_file)
    tiff_files = get_tiff_files(yaml_contents)

    model = get_model(
        existing_model=args.trained_model,
        model_weights=args.model_weights,
        network_depth=models[args.network_depth],
        learning_rate=args.learning_rate,
        continue_training=args.continue_training,
    )

    signal_train, background_train, labels_train = make_lists(tiff_files)

    if args.test_fraction > 0:
        (
            signal_train,
            signal_test,
            background_train,
            background_test,
            labels_train,
            labels_test,
        ) = train_test_split(
            signal_train,
            background_train,
            labels_train,
            test_size=args.test_fraction,
        )
        validation_generator = CubeGeneratorFromDisk(
            signal_test,
            background_test,
            labels=labels_test,
            batch_size=args.batch_size,
            train=True,
        )
    else:
        validation_generator = None

    training_generator = CubeGeneratorFromDisk(
        signal_train,
        background_train,
        labels=labels_train,
        batch_size=args.batch_size,
        shuffle=True,
        train=True,
        augment=not args.no_augment,
    )
    callbacks = []

    if args.tensorboard:
        logdir = output_dir / "tensorboard"
        ensure_directory_exists(logdir)
        tensorboard = TensorBoard(
            log_dir=logdir,
            histogram_freq=0,
            write_graph=True,
            update_freq="epoch",
        )
        callbacks.append(tensorboard)

    if args.save_checkpoints:
        if args.save_weights:
            filepath = str(
                output_dir / "weights.{epoch:02d}-{val_loss:.3f}.h5"
            )
        else:
            filepath = str(output_dir / "model.{epoch:02d}-{val_loss:.3f}.h5")

        checkpoints = ModelCheckpoint(
            filepath,
            save_weights_only=args.save_weights,
            period=args.checkpoint_interval,
        )
        callbacks.append(checkpoints)

    model.fit(
        training_generator,
        validation_data=validation_generator,
        use_multiprocessing=False,
        epochs=args.epochs,
        callbacks=callbacks,
    )

    if args.save_weights:
        print("Saving model weights")
        model.save_weights(str(output_dir / "model_weights.h5"))
    else:
        print("Saving model")
        model.save(output_dir / "model.h5")

    print(
        "Finished training, " "Total time taken: %s",
        datetime.now() - start_time,
    )


if __name__ == "__main__":
    main()



import os
import argparse

import pandas as pd
from brainio import brainio
from imlib.general.numerical import check_positive_float, check_positive_int

from imlib.IO.cells import cells_to_xml
from imlib.source.source_files import get_structures_path
from neuro.structures.IO import load_structures_as_df
from neuro.atlas_tools.misc import get_atlas_pixel_sizes

from cellfinder.tools.prep import prep_atlas_conf
import cellfinder.summarise.count_summary as cells_regions


def xml_crop(args, df_query="name"):
    args = prep_atlas_conf(args)

    if args.structures_file_path is None:
        args.structures_file_path = get_structures_path()

    reference_struct_df = pd.read_csv(get_structures_path())

    curate_struct_df = pd.read_csv(args.structures_file_path)

    curate_struct_df = reference_struct_df[
        reference_struct_df[df_query].isin(curate_struct_df[df_query])
    ]

    curated_ids = list(curate_struct_df["structure_id_path"])

    atlas = brainio.load_any(args.registered_atlas_path)
    hemisphere = brainio.load_any(args.hemispheres_atlas_path)

    structures_reference_df = load_structures_as_df(get_structures_path())

    atlas_pixel_sizes = get_atlas_pixel_sizes(args.atlas_config)
    sample_pixel_sizes = args.x_pixel_um, args.y_pixel_um, args.z_pixel_um

    scales = cells_regions.get_scales(sample_pixel_sizes, atlas_pixel_sizes)

    destination_folder = os.path.join(args.xml_dir, "xml_crop")
    if not os.path.exists(destination_folder):
        os.makedirs(destination_folder)

    xml_names = [f for f in os.listdir(args.xml_dir) if f.endswith(".xml")]
    xml_paths = [os.path.join(args.xml_dir, f) for f in xml_names]

    for idx, xml_path in enumerate(xml_paths):
        print("Curating file: {}".format(xml_names[idx]))
        cells = cells_regions.get_cells_data(
            xml_path, cells_only=args.cells_only,
        )
        max_coords = cells_regions.get_max_coords(cells)

        curated_cells = []
        for i, cell in enumerate(cells):
            cells_regions.transform_cell_coords(atlas, cell, scales)

            structure_id = cells_regions.get_structure_from_coordinates(
                atlas,
                cell,
                max_coords,
                order=args.coordinates_order,
                structures_reference_df=structures_reference_df,
            )
            if structure_id in curated_ids:
                if args.hemisphere_query in [1, 2]:
                    hemisphere = cells_regions.get_structure_from_coordinates(
                        hemisphere,
                        cell,
                        max_coords,
                        order=args.coordinates_order,
                    )
                    if hemisphere is args.hemisphere_query:
                        curated_cells.append(cell)
                else:
                    curated_cells.append(cell)
        cells_to_xml(
            curated_cells,
            os.path.join(destination_folder, xml_names[idx]),
            artifact_keep=True,
        )
    print("Done!")


def get_parser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "--xml-dir",
        dest="xml_dir",
        type=str,
        required=True,
        help="Directory containing XML files to be cropped",
    )

    parser.add_argument(
        "--structures-file",
        dest="structures_file_path",
        type=str,
        help="Curated csv structure list (as per the allen brain atlas csv",
    )

    parser.add_argument(
        "--hemisphere-query",
        dest="hemisphere_query",
        type=int,
        default=0,
        help="Which hemisphere to keep (1 or 2). Default: 0 (all)",
    )

    parser.add_argument(
        "--registered-atlas",
        dest="registered_atlas_path",
        type=str,
        help="The path to the atlas registered to the sample brain",
    )
    parser.add_argument(
        "--hemispheres",
        dest="hemispheres_atlas_path",
        type=str,
        help="The atlas with just the hemispheres encoded.",
    )
    parser.add_argument(
        "--atlas-config",
        dest="atlas_config",
        type=str,
        help="Atlas configuration file. In the same format as the"
        "registration config file",
    )
    parser.add_argument(
        "--cells-only",
        dest="cells_only",
        action="store_false",
        help="Used for testing. Will include non cells in the checks",
    )

    parser.add_argument(
        "--coordinates-order",
        dest="coordinates_order",
        nargs=3,
        type=check_positive_int,
        default=[0, 1, 2],
        help="The order in which to read the dimensions in the atlas from the"
        " cell coordinates. 0,1,2 means x,y,z. 1,0,2 means y,x,z",
    )
    parser.add_argument(
        "-x",
        "--x-pixel-um",
        dest="x_pixel_um",
        type=check_positive_float,
        default=1,
        help="Pixel spacing of the data in the first "
        "dimension, specified in mm.",
    )
    parser.add_argument(
        "-y",
        "--y-pixel-um",
        dest="y_pixel_um",
        type=check_positive_float,
        default=1,
        help="Pixel spacing of the data in the second "
        "dimension, specified in mm.",
    )
    parser.add_argument(
        "-z",
        "--z-pixel-mm",
        dest="z_pixel_um",
        type=check_positive_float,
        default=5,
        help="Pixel spacing of the data in the third "
        "dimension, specified in mm.",
    )
    return parser


def main():
    args = get_parser().parse_args()
    xml_crop(args)


if __name__ == "__main__":
    main()
